@@ -27,7 +27,7 @@
 
     See AdaGenerator.py for an example of use.
 
-    Copyright (c) 2012-2013 European Space Agency
+    Copyright (c) 2012-2020 European Space Agency
 
     Designed and implemented by Maxime Perrotin
 
@@ -35,10 +35,12 @@
 """
 
 import logging
+import operator
+from collections import defaultdict
 LOG = logging.getLogger(__name__)
 
 
-class Expression(object):
+class Expression:
     ''' AST Entry for expressions - Always use subtype '''
     is_raw = False
 
@@ -49,6 +51,8 @@ class Expression(object):
         self.charPositionInLine = charPositionInLine
         # Binary expressions have two sides
         self.right = self.left = None
+        # Unary expressions (NOT and NEG) are stored in "expr"
+        self.expr = None
 
         # exprType is an ASN.1 type (as exported by asn1scc)
         self.exprType = None
@@ -64,14 +68,17 @@ class Expression(object):
 
 class ExprPlus(Expression):
     operand = '+'
+    op = operator.add
 
 
 class ExprMul(Expression):
     operand = '*'
+    op = operator.mul
 
 
 class ExprMinus(Expression):
     operand = '-'
+    op = operator.sub
 
 
 class ExprOr(Expression):
@@ -117,14 +124,17 @@ class ExprLe(Expression):
 
 class ExprDiv(Expression):
     operand = '/'
+    op = operator.truediv  # python3 only
 
 
 class ExprMod(Expression):
     operand = 'mod'
+    op = operator.mod
 
 
 class ExprRem(Expression):
     operand = 'rem'
+    op = operator.mod
 
 
 class ExprNot(Expression):
@@ -134,6 +144,11 @@ class ExprNot(Expression):
 class ExprNeg(Expression):
     operand = '-'
 
+    @property
+    def is_raw(self):
+        ''' Redefine is_raw - check the actual element '''
+        return self.expr.is_raw
+
 
 class ExprAssign(Expression):
     operand = ':='
@@ -159,8 +174,9 @@ class Primary(Expression):
     is_raw = True
 
     def __init__(self, inputString='', line=-1, charPositionInLine=-1,
-                 primary=None):
+                 primary=None, debugLine=-1):
         ''' Initialize common primary attributes '''
+        self.debugLine = debugLine
         if primary:
             self.inputString = primary.inputString
             self.line = primary.line
@@ -220,8 +236,22 @@ class PrimBoolean(Primary):
     pass
 
 
+class PrimNull(Primary):
+    pass
+
+
 class PrimConstant(Primary):
     is_raw = False
+    # other attributes set by the parser:
+    # constant_c_name : string
+    # constant_value : raw value from ASN.1
+
+
+class PrimStateReference(Primary):
+    ''' This primary ("state" keyword) is used only in stop conditions and
+    possibly in the future in observers, to reference the state of the system
+    '''
+    is_raw = False
 
 
 class PrimBitStringLiteral(Primary):
@@ -286,15 +316,14 @@ class PrimSequence(Primary):
     pass
 
 
-class Decision(object):
+class Decision:
     ''' AST Entry for a decision '''
     def __init__(self):
         ''' A DECISION statement '''
         self.inputString = ''
         self.line = None
         self.charPositionInLine = None
-        self.pos_x = 0
-        self.pos_y = 0
+        self.pos_x, self.pos_y = None, None
         self.width = 70
         self.height = 50
         # kind can be 'any', 'informal_text', or 'question'
@@ -318,17 +347,16 @@ class Decision(object):
                 l=self.line, c=self.charPositionInLine)
 
 
-class Answer(object):
+class Answer:
     ''' AST Entry for a decision answer '''
     def __init__(self):
         ''' One ANSWER of a DECISION '''
         self.inputString = ''
         self.line = None
         self.charPositionInLine = None
-        self.pos_x = 0
-        self.pos_y = 0
+        self.pos_x, self.pos_y = None, None
         self.width = 70
-        self.height = 10.5
+        self.height = 23
         # one of 'closed_range' 'constant' 'open_range' 'else' 'informal_text'
         self.kind = None
         # informalText is a string, when kind == 'informal_text'
@@ -341,7 +369,7 @@ class Answer(object):
         # one of ExprEq, ExprNeq, ExprGt, ExprGe, ExprLt, ExprLe (types)
         self.openRangeOp = None
         # transition is of type Transition
-        self.transition = None
+        self.transition : Transition = None
         # optional comment symbol
         self.comment = None
         # optional hyperlink
@@ -353,15 +381,14 @@ class Answer(object):
                 l=self.line, c=self.charPositionInLine)
 
 
-class Task(object):
+class Task:
     ''' AST Entry for TASKS '''
     def __init__(self):
         ''' Initialize TASK attributes (set of ASSIGN statements) '''
         self.inputString = ''
         self.line = None
         self.charPositionInLine = None
-        self.pos_x = 0
-        self.pos_y = 0
+        self.pos_x, self.pos_y = None, None
         self.width = 70
         self.height = 35
         # optional comment symbol
@@ -395,13 +422,12 @@ class TaskForLoop(Task):
     pass
 
 
-class Output(object):
+class Output:
     ''' AST Entry for OUTPUT statements '''
     def __init__(self, defName=''):
         ''' Set of OUTPUT statements '''
         self.inputString = defName
-        self.pos_x = 0
-        self.pos_y = 0
+        self.pos_x, self.pos_y = None, None
         self.width = 70
         self.height = 35
         self.line = None
@@ -420,12 +446,17 @@ class Output(object):
 
 
 class ProcedureCall(Output):
-    ''' Procedure calls content is identical to Outputs '''
-    pass
+    ''' Procedure calls content is identical to Outputs
+    with exception that it may have a return type (exprType) '''
+    def __init__(self, defName=''):
+        super().__init__(defName)
+        # exprType is an ASN.1 type (as exported by asn1scc)
+        self.exprType = None
+        self.is_raw = False
 
 
-class Terminator(object):
-    ''' Terminator elements (join, nextstate, stop) '''
+class Terminator:
+    ''' Terminator elements (join, nextstate, stop, return) '''
     def __init__(self, defName=''):
         ''' Initialize terminator attributes '''
         self.inputString = defName
@@ -450,6 +481,8 @@ class Terminator(object):
         self.via = None
         # 'entrypoint' is the string of the entry point (e.g. "foo")
         self.entrypoint = None
+        # If the nextstate is an instance of a type (foo : bar)
+        self.instance_of = None
         # some transitions can be chained, when entering/leaving nested states
         self.next_id = -1
         # Pointer to the next transition, when using return/connect
@@ -460,6 +493,21 @@ class Terminator(object):
         self.possible_states = []
         # optional composite state content (type CompositeState)
         self.composite = None
+        # Flag to indicate if the nextstate is a state aggregation
+        self.next_is_aggregation = False
+        # List of sibling states if terminator is within a state aggregation
+        # (list of strings with all parallel states of the aggregation)
+        # Used by backend to synchronise parallel state terminations
+        self.siblings = []
+        # If this terminator is within a state aggregation, store the name
+        # of the parallel substate (set by Helper.state_aggregations)
+        self.substate = ''
+        # candidate_id: {transition_id: [states]}
+        # field is set by Helper.py/flatten, in case of "nextstate -"
+        # there is a list of states that set transition_id to -1 : the standard
+        # states ; and there are the composite states, that set a different
+        # id corresponding to the start transition of the state.
+        self.candidate_id = defaultdict(list)
 
     def trace(self):
         ''' Debug output for terminators '''
@@ -469,13 +517,13 @@ class Terminator(object):
                 x=self.pos_x, y=self.pos_y)
 
 
-class Label(object):
+class Label:
     ''' AST Entry for a Label '''
     def __init__(self):
         ''' Initialize the label attributes '''
         # inputString holds the label name
         self.inputString = ''
-        self.pos_x = 0
+        self.pos_x, self.pos_y = None, None
         self.pos_y = 0
         self.width = 70
         self.height = 35
@@ -496,7 +544,7 @@ class Label(object):
                 l=self.line, c=self.charPositionInLine)
 
 
-class Floating_label(Label, object):
+class Floating_label(Label):
     ''' AST Entry for a floating label '''
     def __init__(self, label=None):
         ''' Initialize the floating label attributes '''
@@ -518,7 +566,7 @@ class Floating_label(Label, object):
                 l=self.line, c=self.charPositionInLine)
 
 
-class Transition(object):
+class Transition:
     ''' AST Entry for a complete transition '''
     def __init__(self):
         ''' Initialize the transition attributes '''
@@ -528,7 +576,7 @@ class Transition(object):
         self.terminator = None
         # All Terminators of this transition
         self.terminators = []
-        # List of State that can lead to this transition
+        # List of states (string) that can lead to this transition
         # There can be several if state has multiple names (e.g. STATE a, b)
         # Note, this field is updated by the Helper.flatten function
         # and is needed to properly know when to call a nested state exit
@@ -543,14 +591,13 @@ class Transition(object):
         return u'\n'.join(data)
 
 
-class Input(object):
+class Input:
     ''' AST Entry for the INPUT symbol '''
     def __init__(self):
         ''' Initialize the Input attributes '''
         # inputString is the user text, it can contain several inputs
         self.inputString = ''
-        self.pos_x = 0
-        self.pos_y = 0
+        self.pos_x, self.pos_y = None, None
         self.width = 70
         self.height = 35
         self.line = None
@@ -593,14 +640,28 @@ class Connect(Input):
         return u'CONNECT {exp} ({l},{c})'.format(exp=self.inputString,
                 l=self.line, c=self.charPositionInLine)
 
+class ContinuousSignal(Input):
+    ''' AST Entry for the Continuous Signal '''
+    def __init__(self):
+        ''' Difference with Input: trigger is an expression '''
+        super(ContinuousSignal, self).__init__()
+        # Decision triggering the transition
+        self.trigger = None
+        # Priority (integer)
+        self.priority = 0
 
-class Start(object):
+    def trace(self):
+        ''' Debug output for a Continuous signal '''
+        return u'PROVIDED {exp} ({l},{c})'.format(exp=self.inputString,
+                l=self.line, c=self.charPositionInLine)
+
+
+class Start:
     ''' AST Entry for the START symbol '''
     def __init__(self):
         ''' Initialize the Start symbol attributes '''
         self.inputString = ''
-        self.pos_x = 0
-        self.pos_y = 0
+        self.pos_x, self.pos_y = None, None
         self.width = 70
         self.height = 35
         # start transition is of type Transition
@@ -627,14 +688,14 @@ class CompositeState_start(Start):
     pass
 
 
-class Comment(object):
+class Comment:
     ''' AST Entry for COMMENT symbols '''
     def __init__(self):
         ''' Comment symbol '''
         # inputString is the comment value itself
         self.inputString = ''
-        self.pos_x = 0
-        self.pos_y = 0
+        self.pos_x = None
+        self.pos_y = None
         self.width = 70
         self.height = 35
         self.line = None
@@ -649,7 +710,7 @@ class Comment(object):
                 l=self.line, c=self.charPositionInLine)
 
 
-class State(object):
+class State:
     ''' AST Entry for STATE symbols '''
     def __init__(self, defName=''):
         ''' Used only for rendering backends - not for code generation '''
@@ -659,7 +720,7 @@ class State(object):
         self.statelist = []
         self.line = None
         self.charPositionInLine = None
-        self.pos_x = 0
+        self.pos_x, self.pos_y = None, None
         self.pos_y = 0
         self.width = 70
         self.height = 35
@@ -667,12 +728,19 @@ class State(object):
         self.inputs = []
         # list of type Connect (connection below a nested state)
         self.connects = []
+        # list of ContinuousSignal (provided clauses below a state)
+        self.continuous_signals = []
         # optional comment symbol
         self.comment = None
         # optional hyperlink
         self.hyperlink = None
         # optional composite state content (type CompositeState)
         self.composite = None
+        # via clause, used for entering nested state with an entry point
+        # 'via' is the string for the renderer (e.g. "hello via foo")
+        self.via = None
+        # a state can be an instance of a state stype
+        self.instance_of = None
 
     def trace(self):
         ''' Debug output for a STATE symbol '''
@@ -681,17 +749,15 @@ class State(object):
                 x=self.pos_x, y=self.pos_y)
 
 
-class TextArea(object):
+class TextArea:
     ''' AST Entry for text areas (containing declarations/comments) '''
     def __init__(self):
         ''' Text area (raw content for rendering only) '''
-        self.inputString = '-- Declare your variables\n\n' \
-                           '-- Syntax: DCL <variable name> <type name>;\n\n'
+        self.inputString = '-- Text area for declarations and comments\n\n'
         self.line = None
         self.charPositionInLine = None
         # Set default coordinates and width/height
-        self.pos_x = 0
-        self.pos_y = 0
+        self.pos_x, self.pos_y = None, None
         self.width = 170
         self.height = 140
         # DCL variables in the text area {name: (sort, default_value), ...}
@@ -699,6 +765,12 @@ class TextArea(object):
         # fpar and timers are also listed, useful when using autocompletion
         self.fpar = []
         self.timers = []
+        # signals can be declared in a text area (structure: see System class)
+        self.signals = []
+        # USE construct can be declared in a text area (list of strings)
+        # asn1_files lists optional COMMENT strings of USE constructs
+        self.use_clauses = []
+        self.asn1_files = []
         # inner procedures - useful to get autocompletion params
         self.procedures = []
         # optional hyperlink
@@ -710,8 +782,8 @@ class TextArea(object):
                 l=self.line, c=self.charPositionInLine)
 
 
-class Automaton(object):
-    ''' Elements contained in a process or a procedure '''
+class Automaton:
+    ''' Elements contained in a process, procedure or composite state'''
     def __init__(self, parent=None):
         ''' AST grouping the elements that can be rendered graphically '''
         self.parent = parent
@@ -723,7 +795,7 @@ class Automaton(object):
         self.named_start = []
 
 
-class Procedure(object):
+class Procedure:
     ''' Internal procedure definition '''
     def __init__(self):
         ''' Procedure AST default value '''
@@ -731,7 +803,7 @@ class Procedure(object):
         self.line = None
         self.charPositionInLine = None
         # Set default coordinates and width/height
-        self.pos_x = self.pos_y = 0
+        self.pos_x = self.pos_y = None
         self.width = 70
         self.height = 35
         # Optional hyperlink
@@ -747,6 +819,8 @@ class Procedure(object):
         self.fpar = []
         # return type (ASN.1)
         self.return_type = None
+        # when procedure has a RETURN it can also contain a variable name
+        self.return_var = None
         # start, terminators, text areas, floating_labels (see Process)
         #self.start = None
         #self.states = []
@@ -766,6 +840,8 @@ class Procedure(object):
         self.transitions = []
         # Determine if a procedure is externally defined
         self.external = False
+        # Determine if a procedure is a remote procedure
+        self.exported = False
         # Optional comment
         self.comment = None
         # Set of symbols contained in the procedure (type Automaton)
@@ -773,15 +849,22 @@ class Procedure(object):
         # input/output signal lists - unused but for context information
         self.input_signals = []
         self.output_signals = []
+        # The "DECISION ANY" construct requires random number generators
+        self.random_generator = set()
 
 
-class Process(object):
+class Process:
     ''' SDL Process entry point '''
     def __init__(self):
         ''' Process AST default values '''
         self.processName = None
         # Optional filename containing this process (PR file)
         self.filename = None
+        # Optional type of this process instance (name = string, ref = Process)
+        self.instance_of_name = None
+        self.instance_of_ref  = None
+        # A process may be a process type (boolean)
+        self.process_type = False
         # Process parent context (Block or AST if defined at root level)
         self.parent = None
         # A process can be referenced (externally defined)
@@ -804,6 +887,10 @@ class Process(object):
         # dataview: complete AST of the ASN.1 types
         self.asn1Modules = None
         self.dataview = None
+        # Reference to the Python module containing the ASN.1 AST
+        self.DV = None
+        # user-defined datatypes (using "newtype" SDL keyword)
+        self.user_defined_types = dict()
 
         # input and output signal lists:
         # [{'name': str, 'type': str, 'direction':'in'/'out'}]
@@ -827,6 +914,9 @@ class Process(object):
         # then Input contains the inputs list and corresponding transition
         self.mapping = {}
 
+        # Similar mapping for continuous signals
+        self.cs_mapping = defaultdict(list)
+
         # list of type Transition - use 'mapping' to map index to inputs/states
         self.transitions = []
 
@@ -837,9 +927,19 @@ class Process(object):
         # (Includes inner procedures)
         self.content = Automaton(parent=self)
 
+        # Process formal parameters - list of dict:
+        # [{'name': str, 'type': str}]
+        self.fpar = []
+
         # List of timers (strings) declared in the process
         self.timers = []
 
+        # list of processes used for context management of the GUI
+        self.processes = []
+
+        # The "DECISION ANY" construct requires random number generators
+        self.random_generator = set()
+
 
 class CompositeState(Process):
     '''
@@ -851,8 +951,8 @@ class CompositeState(Process):
     def __init__(self):
         super(CompositeState, self).__init__()
         self.statename = ''
-        self.state_entrypoints = []
-        self.state_exitpoints = []
+        self.state_entrypoints = set()
+        self.state_exitpoints = set()
         # Special entry and exit procedures (named "entry" and "exit")
         self.entry_procedure = None
         self.exit_procedure = None
@@ -866,7 +966,29 @@ class CompositeState(Process):
                 l=self.line, c=self.charPositionInLine)
 
 
-class Block(object):
+class StateAggregation(CompositeState):
+    '''
+        State Aggregation (Parallel states) are supported since SDL2000
+        These states can only contain (in the self.content field):
+            text areas
+            procedure definitions
+            composite states (including sub-state aggregations)
+        But no state machine definition
+    '''
+    def __init__(self):
+        super(StateAggregation, self).__init__()
+        # List of partition connections:
+        # [{'outer': {'state_part_id': str, 'point': str},
+        #   'inner': {'state_part_id': str, 'point': str}}]
+        self.state_partition_connections = []
+
+    def trace(self):
+        ''' Debug output for state aggregation '''
+        return u'STATE AGGREGATION {exp} ({l},{c})'.format(exp=self.statename,
+                l=self.line, c=self.charPositionInLine)
+
+
+class Block:
     ''' AST for a BLOCK entity '''
     def __init__(self):
         '''
@@ -887,11 +1009,17 @@ class Block(object):
         self.connections = []
         # list of class Block
         self.blocks = []
-        # list of class Process
+        # list of class Process / with separate process_types
         self.processes = []
+        self.process_types = []
+        # list of ogAST.Procedure
+        self.procedures = []
+        # Block formal parameters - list of dict (unused)
+        # [{'name': str, 'type': str}]
+        self.fpar = []
 
 
-class System(object):
+class System:
     ''' AST for a SYSTEM entity '''
     def __init__(self):
         '''
@@ -904,7 +1032,7 @@ class System(object):
         self.filename = None
         # Reference to top-level AST
         self.ast = None
-        # list of SIGNAL declarations: [{'name': str, 'type': str}]
+        # list of SIGNAL declarations: [{'name': str, 'type': asn1type}]
         # (Supporting only one parameter)
         self.signals = []
         # list of ogAST.Procedure
@@ -913,9 +1041,10 @@ class System(object):
         # 'signals': ['sig1', .. ]}]
         self.channels = []
         self.blocks = []
+        self.text_areas = []
 
 
-class AST(object):
+class AST:
     ''' Top-level AST entry point '''
     def __init__(self):
         '''
@@ -931,11 +1060,16 @@ class AST(object):
         # Refs to the ASN.1 dataview AST (set with USE clauses)
         self.dataview = None
         self.asn1Modules = None
+        # DV is the Asn1scc imported module
+        self.DV = None
         # ASN.1-defined constants (constants in Ada but variables in C)
         # dictionnary: {ConstantName: type } - copied from dataview.py
-        self.asn1_constants = None
+        self.asn1_constants = {}
 
         # List of System
         self.systems = []
         # List of Process
         self.processes = []
+
+        # List of Process types
+        self.process_types = []
