@@ -1,8 +1,7 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
-
 """
-    OpenGEODE - A tiny SDL Editor for TASTE
+    OpenGEODE - A tiny SDL Editor for TASTE - Support for Undo/Redo
 
     Undo/Redo commands for generic symbols when used in a diagram editor.
 
@@ -10,7 +9,7 @@
         the redo() function is *called* when the command is created.
         No need to perform the action before.
 
-    Copyright (c) 2012 European Space Agency
+    Copyright (c) 2012-2020 European Space Agency
 
     Designed and implemented by Maxime Perrotin
 
@@ -18,14 +17,18 @@
 """
 
 
-from PySide.QtGui import QUndoCommand
-from PySide.QtCore import QPropertyAnimation, QEasingCurve, QAbstractAnimation
-
 import logging
+from PySide2.QtGui import *
+from PySide2.QtCore import *
+from PySide2.QtWidgets import *
+
+from . import Pr
+from . import ogParser
+
 LOG = logging.getLogger(__name__)
 
 
-class UndoMacro(object):
+class UndoMacro:
     ''' Context manager for Undo macros '''
 
     def __init__(self, stack, text):
@@ -35,12 +38,10 @@ class UndoMacro(object):
 
     def __enter__(self):
         ''' Create macro for a set of undo commands '''
-        LOG.debug('New macro: {}'.format(self.text))
         self.stack.beginMacro(self.text)
 
     def __exit__(self, atype, value, traceback):
         ''' Stop the recording of undo commands '''
-        LOG.debug('End macro: {}'.format(self.text))
         self.stack.endMacro()
 
 
@@ -48,24 +49,90 @@ class ReplaceText(QUndoCommand):
     ''' Undo/Redo command for updating the text in a symbol '''
     def __init__(self, text_id, old_text, new_text):
         super(ReplaceText, self).__init__()
-        LOG.debug('New Undo command: Replace text')
         self.setText('Replace text')
         self.text = text_id
         self.old_text = old_text
         self.new_text = new_text
+        self.scene = self.text.parent.scene()
+        self.count_instances_old, self.count_instances_new = 1, -1
+        for each in self.scene.states:
+            # count the number of instances of the states in the scene
+            # the text is already renamed, values are initialized to -1 and 1
+            state_name = str(each).lower()
+            if self.old_text.lower() == state_name:
+                self.count_instances_old += 1
+            if self.new_text.lower() == state_name:
+                self.count_instances_new += 1
 
     def undo(self):
         self.text.setPlainText(self.old_text)
+        if self.text.parent in self.scene.states:
+            # see explanations in the redo() function below
+            if self.count_instances_new != 0:
+                pass
+            else:
+                if self.count_instances_old == 1:
+                    try:
+                        self.scene.composite_states[self.old_text.lower()] = \
+                         self.scene.composite_states.pop(self.new_text.lower())
+                    except KeyError:
+                        pass
+                else:
+                    # scene was created with redo, delete reference
+                    try:
+                        self.scene.composite_states.pop(self.new_text.lower())
+                    except KeyError:
+                        pass
 
     def redo(self):
         self.text.setPlainText(self.new_text)
 
+        if self.text.parent in self.scene.states:
+            # renaming a state in case of a nested state:
+            # 1) if renamed state already exists, do nothing special
+            # 2) if renamed state is a new state:
+            #    a. if this was the only instance of the state, just rename
+            #       the list of composite state to reflect the new state name
+            #    b. otherwise, parse the composite state, create a new scene
+            #       and render the same content in the new scene, then
+            #       update the list of composite states with the new copy
+            if self.count_instances_new != 0:
+                # case 1: do nothing, state is already initialized
+                pass
+            else:
+                if self.count_instances_old == 1:
+                    # case 2a. rename the list of composite states
+                    try:
+                        self.scene.composite_states[self.new_text.lower()] = \
+                         self.scene.composite_states.pop(self.old_text.lower())
+                    except KeyError:
+                        pass
+                else:
+                    # case 2b: create a new scene and really copy the state
+                    state = self.text.parent
+                    if state.is_composite():
+                        # Parse the scene of the nested state:
+                        sub_state = u'\n'.join(Pr.generate(state,
+                                                           composite=True,
+                                                           nextstate=False))
+                        if sub_state:
+                            new_scene = self.scene.create_subscene(
+                                    context=state.context_name,
+                                    parent=self.scene)
+                            # get the AST of type ogAST.CompositeState
+                            ast, _, _, _, _ = ogParser.parseSingleElement(
+                                    elem='composite_state',
+                                    string=sub_state)
+                            # render the composite state content
+                            new_scene.render_everything(ast.content)
+                            self.scene.composite_states[
+                                    self.new_text.lower()] = new_scene
+
 
 class ResizeSymbol(QUndoCommand):
     ''' Undo/Redo command for resizing a symbol '''
     def __init__(self, symbol_id, old_rect, new_rect):
         super(ResizeSymbol, self).__init__()
-        LOG.debug('New Undo command: Resize symbol')
         self.setText('Resize symbol')
         self.symbol = symbol_id
         self.old_rect = old_rect
@@ -94,7 +161,6 @@ class InsertSymbol(QUndoCommand):
     ''' Undo/Redo command for inserting a new item '''
     def __init__(self, item, parent, pos):
         super(InsertSymbol, self).__init__()
-        LOG.debug('New Undo command: Insert symbol')
         self.item = item
         self.parent = parent
         self.pos_x = pos.x() if pos else None
@@ -122,7 +188,11 @@ class InsertSymbol(QUndoCommand):
         self.item.show()
         self.item.grabber.display()
         try:
-            self.scene.refresh()
+            # Immediate refresh for a better rendering, howerver slows down
+            # performance when inserting decisions in a complex model
+            #self.scene.refresh()
+            # Update only the connections - this is much faster
+            self.item.update_connections()
         except AttributeError:
             pass
 
@@ -131,7 +201,6 @@ class DeleteSymbol(QUndoCommand):
     ''' Undo/Redo command for a symbol deletion '''
     def __init__(self, item):
         super(DeleteSymbol, self).__init__()
-        LOG.debug('New Undo command: Delete symbol')
         self.item = item
         self.scene = item.scene()
         self.parent = item.parentItem() if item.hasParent else None
@@ -159,22 +228,20 @@ class MoveSymbol(QUndoCommand):
     ''' Undo/Redo command for moving symbols '''
     def __init__(self, symbol_id, old_pos, new_pos, animate=False):
         super(MoveSymbol, self).__init__()
-        LOG.debug('New Undo command: Move symbol')
         self.setText('Move symbol')
         self.symbol = symbol_id
         self.old_pos = old_pos
         self.new_pos = new_pos
         if animate:
-            self.animation = QPropertyAnimation(self.symbol, "position")
+            self.animation = QPropertyAnimation(self.symbol, b"position")
             self.animation.setDuration(500)
             self.animation.setStartValue(self.old_pos)
             self.animation.setEndValue(self.new_pos)
             self.animation.setEasingCurve(QEasingCurve.OutCirc)
 
-
     def undo(self):
         ''' Undo a symbol move '''
-        self.symbol.setPos(self.old_pos)
+        self.symbol.position = self.old_pos
         try:
             self.symbol.decisionParent.updateConnectionPointPosition()
         except AttributeError:
@@ -185,7 +252,7 @@ class MoveSymbol(QUndoCommand):
         try:
             self.animation.start()
         except AttributeError:
-            self.symbol.setPos(self.new_pos)
+            self.symbol.position = self.new_pos
         try:
             self.symbol.decisionParent.updateConnectionPointPosition()
         except AttributeError:
